/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Googlebar for Mozilla.
 *
 * The Initial Developer of the Original Code is Andy Edmonds.
 * Portions created by the Initial Developer are Copyright (C) 2001
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *  Andy Boughton
 *  Andrew Houghton
 *  John Woods
 *  Bernd Kuemmerlen
 *  Robert Mulcahy
 *  Alfred Kayser
 *  Reflex
 *  Gary Turnbull
 *  Raj Bhaskar
 *  Joachim Thewes
 *  Henrik Gemal
 *  Robert Fernandes
 *  Joe Chellman
 *  Franki Cheung
 *  Alban Fonrouge
 *  Martin Hassman
 *  Ufuk Kayserilioglu
 *  Yoni Gilad
 *  Tim Schmidt
 *  timeless
 *  Francis Turner
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

try {
	const GB_CONSOLE_SVC = Components.classes['@mozilla.org/consoleservice;1'].getService(Components.interfaces.nsIConsoleService);
}
catch (e) {
	myGooglebarUtil.logMessage("Failed to initialize console service\n" + e);
}

try {
	const GB_IO_SVC = Components.classes["@mozilla.org/network/io-service;1"].getService(Components.interfaces.nsIIOService);
}
catch (e) {
	myGooglebarUtil.logMessage("Failed to initialize IO service\n" + e);
}

try {
	const GB_PROMPT_SVC = Components.classes["@mozilla.org/embedcomp/prompt-service;1"].getService(Components.interfaces.nsIPromptService);
}
catch (e) {
	myGooglebarUtil.logMessage("Failed to initialize prompt service\n" + e);
}

try {
	const GB_WINDOW_SVC = Components.classes["@mozilla.org/appshell/window-mediator;1"].getService(Components.interfaces.nsIWindowMediator);
}
catch (e) {
	myGooglebarUtil.logMessage("Failed to initialize window service\n" + e);
}

try {
	const GB_DIR_OBJ = Components.classes['@mozilla.org/file/directory_service;1'].createInstance().QueryInterface(Components.interfaces.nsIProperties);
}
catch (e) {
	myGooglebarUtil.logMessage("Failed to initialize directory object\n" + e);
}

try {
	const GB_FIND_OBJ = Components.classes["@mozilla.org/embedcomp/find;1"].createInstance().QueryInterface(Components.interfaces.nsIWebBrowserFindInFrames);
}
catch (e) {
	myGooglebarUtil.logMessage("Failed to initialize find object\n" + e);
}

try {
	const GB_SOUND_OBJ = Components.classes["@mozilla.org/sound;1"].createInstance().QueryInterface(Components.interfaces.nsISound);
}
catch (e) {
	myGooglebarUtil.logMessage("Failed to initialize sound object\n" + e);
}

try {
	const GB_URL_OBJ = Components.classes["@mozilla.org/network/standard-url;1"].createInstance().QueryInterface(Components.interfaces.nsIURL);
}
catch (e) {
	myGooglebarUtil.logMessage("Failed to initialize URL object\n" + e);
}

const GB_DEBUG_DUMP = false;	// Enable to see dump output - see logMessage function
const GB_DEBUG_MSG  = true;		// Enable to see logs in js console message secion - see logMessage function


var myGooglebarUtil = {

	/*** Variable Declarations ***/
	
	mInFirefox: false,
	mInPrefs: false,

	/*** Function Declarations ***/
	
	arrayIndexOf: function(aItem, aArray) {
		if (!aArray || aArray.length <= 0) {
			return -1;
		}
			
		for (var i = 0; i < aArray.length; i++) {
			if (aArray[i] == aItem) {
				return i;
			}
		}
		
		return -1;
	},
	
	buildURI: function(aCriteria) {
	    var terms = aCriteria.split(" ");
	    var uri = "";
	
	    for (var i = 0; i < terms.length; i++) {
	        if (i > 0) {
	            uri += "+";
	        }
	        uri += this.escapeTerm(terms[i]);
	    }
	    return uri;
	},

	buildURL: function(aPathArray, aOffset) {
		var mystr = '';

		for (var i = 0; i < aOffset; i++) {
			if (aPathArray[i].length) {
				mystr += aPathArray[i];
				mystr += "/";
	            
				if (i < (aOffset - 1)) {
					if (i == 0) {                     
						if (aPathArray[i] == "file:") {
							mystr += "/";
						}

						// if it's the first item (protocol) then add a second /
						if (i == 0) {
							mystr += "/";
						}
					}
				}
			}
		}
	
	    return mystr;
	},

	clearSelectedText: function() {
		try {
			goDoCommand('cmd_selectNone');
		}
		catch (e) {
			this.logMessage("Problem clearning selected text.\n" + e);
		}
	},

	escapeTerm: function(aTerm) {
	    // we need to handle both '/' and '+' specially and '.'
	    aTerm = encodeURIComponent(aTerm);
	    return aTerm.replace(/\+/g,"%2B").replace(/\//g,"%2F").replace(/\./g,"%2E");
	},

	findInPage: function(aEvent, aString) {
		/* This is shamelessly taken from Firebird findUtils.js */
		var found = false;
		var focusedWindow = document.commandDispatcher.focusedWindow;

		if (!focusedWindow || focusedWindow == window) {
			focusedWindow = window._content;
		}
		
		var findInst = getBrowser().webBrowserFind;
		var findInFrames = findInst.QueryInterface(Components.interfaces.nsIWebBrowserFindInFrames);
		findInFrames.rootSearchFrame = window._content;
		findInFrames.currentSearchFrame = focusedWindow;
	
		// setup the find instance
		findInst.searchString  = aString;
	
		if (aEvent) {
			// we hardcode the search options here with what seems to make most sense
			findInst.searchFrames  = true;
			findInst.matchCase     = aEvent.ctrlKey;
			findInst.wrapFind      = true;
			findInst.entireWord    = aEvent.altKey;
			findInst.findBackwards = aEvent.shiftKey;
	  
	  		if (findInst.entireWord) {
				findInst.searchString  = ' ' + aString + ' ';
	  		}
		}

		try {
			found = findInst.findNext();
		}
		catch (e) {
		}
	
		if (!found && GB_SOUND_OBJ) {
			GB_SOUND_OBJ.beep();
			// alert('"' + aString + '" ' + this.getProperty("gb.notfound"));
		}
	},

	findTerms: function(aCriteria) {
	    //  quotes only matter when preceded by a space or a quote.
	    var terms    = new Array();
	    var inQuote  = false;
	    var haveTerm = false;

	    var currPtr  = 0;    // the start of the current term
	    var prevChar = ' ';  // the char we last saw

	    var val;
		aCriteria = aCriteria.replace(/(cache|link|related|stocks|define|site|allintitle|intitle|allinurl|inurl|filetype|group):/g, "");
		
	    for (var index = 0; index < aCriteria.length; index++) {
	        var currChar = aCriteria.charAt(index);
	
			switch (currChar) {
	          
				case '+':
					if (!inQuote && !haveTerm) {
						currPtr = index;
					}
					break;
				
				case '.':				
					if (!inQuote) {
						val = this.trimString(aCriteria.substring(currPtr, index));
						
						if (val.indexOf('http://') == 0 || val.indexOf('https://') == 0 || val.indexOf('ftp://') == 0) {
							haveTerm = true;
							break;
						}
						else if (/^\d+(\.\d+)*$/i.test(val)) {
							haveTerm = true;
							break;
						}
					}
				
				case ' ':
				case '(':
				case ')':
				case '~':
				case ',':
				// double-byte space
				case decodeURIComponent('%E3%80%80'):
					// these characters do not occur on the search term buttons, 
					// except in quoted phrases
					if (!inQuote && haveTerm) {
						val = this.trimString(aCriteria.substring(currPtr, index));

						if (val[0] != '-' && val != 'OR' && val != 'AND' && !this.inArray(terms, val.replace(/\-/g, ""))) {
							terms.push(val.replace(/\-/g, " "));
							currPtr = index;
						}
						haveTerm = false;						
					}
					
					break;

				case '"':
					// phrases (enclosed in "") should result in only one button
					if (!inQuote) {
						if (haveTerm) {
							val = this.trimString(aCriteria.substring(currPtr, index));

							if (!this.inArray(terms, val)) {
								terms.push(val.replace(/\-/g, " "));
							}
							haveTerm = false;
						}
	
						// any of the discardable characters can occur before a phrase

						if ((prevChar == '+')
							|| (prevChar == ' ')
							|| (prevChar == '(')
							|| (prevChar == ')')
							|| (prevChar == '+')
							|| (prevChar == '~')
							|| (prevChar == ',')
							|| (prevChar == '"'))
						{

							inQuote = true;
						}
					}
					else {
						if (haveTerm) {
							val = aCriteria.substring(currPtr, index);

							if (!this.inArray(terms, val)) {
								terms.push(val);
							}
							haveTerm = false;
							inQuote = false;
						}
					}
					break;

	        	default:
					if (!haveTerm) {
						haveTerm = true;
						currPtr = index;
					}
			}
			prevChar = currChar;
		}
		
		val = this.trimString(aCriteria.substring(currPtr, index));

		if (haveTerm && val[0] != '-' && val != 'OR' && val != 'AND' && !this.inArray(terms, val.replace(/\-/g, ""))) {
			terms.push(val.replace(/\-/g, " "));
		}
	    
	    return terms;
	},

	getAnchorURL: function(aElement, aBaseURL) {
	    var href = aElement.getAttribute("href");
	    var url;
	
	    if (!href || /^\s*mailto:/i.test(href)) {
	        return null;
	    }
	    else {
	        GB_URL_OBJ.spec = aBaseURL;
	        url = GB_URL_OBJ.resolve(href);
	    }
	
	    if (url) {
	        return url;
	    }
	    else {
			alert(this.getProperty("gb.getAnchorURL.1") + "\n" + href);
	        return null;
	    }
	},
	
	/*
		A convenience function to get googlebar elements regardless of where
		the function is being called from within the googlebar framework
		(toolbar, popups, dialogs, prefs, etc.)  Uses the inPrefs flag to
		attempt to find the element from opener first, otherwise it attempts
		to first use the current document.
	*/
	getAnonymousElement: function(aParent, aAttr, aValue) {
		if (!aParent || !aAttr || !aValue) { return null; }
		
		var doc, element;
		
		try {
			doc = this.isInPrefs() ? top.opener.document : document;
			element = doc.getAnonymousElementByAttribute(aParent, aAttr, aValue);
		}
		catch (e) {
			element = null;
		}
		
		if (element != null) { return element; }
		
		try {
			doc = this.isInPrefs() ? document : top.opener.document;
			element = doc.getAnonymousElementByAttribute(aParent, aAttr, aValue);
		}
		catch (e) {
			element = null;
		}
		
		return element;
	},

	/*
		A convenience method which uses the getElement function
		to retrieve an element from anywhere within the googlebar
		framework (toolbar, popups, dialogs, prefs, etc.) and
		get an attribute value from that element.
	*/
	getAttribute: function(aId, aAttr) {
		if (!aId || !aAttr) { return null; }
		
		var element = this.getElement(aId);
		
		if (element == null) { return null; }
	
		return element.getAttribute(aAttr);
	},

	/*
		A convenience function to get googlebar elements regardless of where
		the function is being called from within the googlebar framework
		(toolbar, popups, dialogs, prefs, etc.)  Use the inPrefs flag to
		attempt to find the element from opener first, otherwise it attempts
		to first use the current document.
	*/
	getElement: function(aID) {
		if (!aID) { return null; }
		
		var doc, element;
		
		try {
			doc = this.isInPrefs() ? top.opener.document : document;
			element = doc.getElementById(aID);
		}
		catch (e) {
			element = null;
		}
		
		if (element != null) { return element; }
	
		try {
			doc = this.isInPrefs() ? document : top.opener.document;
			element = doc.getElementById(aID);
		}
		catch (e) {
			element = null;
		}
		
		// If prefs window was opened from the fx extensions
		// panel then we may need to look for a document with
		// the googlebar contained within it
		if (this.isInPrefs() && element == null) {
			this.logMessage("looking for browser window");
			var browserWindow = GB_WINDOW_SVC.getMostRecentWindow("navigator:browser");
			if (browserWindow) {
				this.logMessage("looking for element in browser window");
				element = browserWindow.document.getElementById(aID);
			}
		}
		
		return element;
	},

	getFileURI: function(aFilePath, aExists) {
		var newURI;
		
		// Interface change - ns7 (mozilla 1.0) needs nsIIOService
		// latter (mozilla 1.2) needs nsIFileProtocolHandler
		if (Components.interfaces.nsIFileProtocolHandler) {
			var fileHandler = GB_IO_SVC.getProtocolHandler("file").QueryInterface(Components.interfaces.nsIFileProtocolHandler);  
			var file = Components.classes["@mozilla.org/file/local;1"].createInstance(Components.interfaces.nsILocalFile);
			file.initWithPath(aFilePath);
			
			if (aExists) {
				newURI = file.exists() ? fileHandler.getURLSpecFromFile(file) : "";
			}
			else {
				newURI = fileHandler.getURLSpecFromFile(file);
			}
		}
		else {
			newURI = GB_IO_SVC.getURLSpecFromFile(aFilePath);
		}
		    
		return newURI;
	},
	
	getProperty: function(aKey) {
		try {
			return this.getElement("bundle_googlebar").getString(aKey);
		}
		catch (e) {
			this.logMessage("Problem retrieving property \"" + aKey + "\": " + e);
			return null;
		}
	},
	
	getSelectedText: function(aIgnoreLoc) {
	    var textBox;
	    var text;
	
	    // attempt to grab selected text from the current document (tab)
	    // try to use new window.getSelection() function first,
	    // failover to old document.getSelection()
	    try {
	    	text = new String(window.getSelection());
	    }
	    catch(e) {
	    	this.logMessage("window.getSelection() is not valid, trying document.getSelection()");	    
	    }

	    if (!text || text.length == 0) {
	    	try {
	    		text = new String(window._content.getSelection());
	    	}
	    	catch (e) {
	    		this.logMessage("document.getSelection() is not happy (1st attempt)\n" + e);
	    	}
		}
	
	    // if nothing selected in the document then perhaps there is a
	    // selection contained within a frame of the current document
	    if (!text || text.length == 0) {
	        for (var i = 0; i < _content.frames.length; i++) {
	        	try {
		            text = new String(_content.frames[i].document.getSelection());
		        }
		        catch (e) {
		        	this.logMessage("document.getSelection() is not happy (2nd attempt)\n" + e);
		        }
	            if (text) break;
	        }
	    }
	
	    // if still no text has been found then perhaps something in
	    // the the location text box has been selected
	    if ((!text || text.length == 0) && !aIgnoreLoc) {
	        textBox = window.document.getElementById("urlbar");
	        if (textBox.selectionStart >= 0) {
	            text = textBox.value.substring(textBox.selectionStart, textBox.selectionEnd);
	        }
	    }
	
	    // if still no text has been found then perhaps something in
	    // the the criteria box has been selected - not sure why you
	    // would want to search on something already in the criteria
	    // box but hey, who am I to criticize
	    // hmmm, here is a far fetched use case so I'll leave this code
	    // for now - someone may have multiple terms in the criteria
	    // box and may want to only search on one of them, highlighting
	    // that single term and then shift+click on highlight search
	    // button would search on that term alone - not likely, but a use case
	    if (!text || text.length == 0) {
	        textBox = this.getElement("ggb_search_criteria");
	        if (textBox.selectionStart >= 0) {
	            text = textBox.value.substring(textBox.selectionStart, textBox.selectionEnd);
	        }
	    }

		// Still NO text!!!
		// Now we'll look for forms input boxes
		// yes, try and get the text
		// no, continue on with empty search
		if (!text || text.length == 0) {
			var formsArray = _content.document.forms;
			if (!formsArray) {
				return "";
			}
	
			var form;
			for (form = 0; form < formsArray.length; form++) {
				var elementsArray = formsArray[form].elements;
				var element;
	
				for (element = 0; element < elementsArray.length; element++) {
					var input = elementsArray[element];
	           
					if ((input.type == "text") || (input.type == "textarea")) {
						// For some reason news.google.com, and potentially other sites
						// cause this input.selectionStart to blow up in the browser
						// with a javascript no data error.  Wrapping for safety.
						try {
							if (input.selectionStart != input.selectionEnd) {
								text = input.value.substring(input.selectionStart, input.selectionEnd);
								break;
							}
						}
						catch(e) {
							// Munch and ignore error
						}
					}
				}
			}
		}

		return text;
	},
	
	inArray: function(aArray, aString) {
		for (var i = 0; i < aArray.length; i++) {
			if (aArray[i].toLowerCase() == aString.toLowerCase()) {
				return true;
			}
		}
		return false;
	},

	initialize: function() {
		// TODO: Someday find a real way to do this besides looking for a null broadcaster
		this.mInFirefox = this.getElement('ggb_contextpref') == null;
		
		if (GB_SOUND_OBJ) {
			GB_SOUND_OBJ.init();
		}
	},

	isInFirefox: function() {
		return this.mInFirefox;
	},
	
	isInPrefs: function() {
		return this.mInPrefs;
	},
	
	logMessage: function(aMessage) {
		if (GB_DEBUG_DUMP) { dump('Googlebar: ' + aMessage); }
		if (GB_DEBUG_MSG) { GB_CONSOLE_SVC.logStringMessage('Googlebar: ' + aMessage); }
	},
	
	normalizeString: function(aString) {
	    // trim leading and trailing spaces
	    var newString = this.trimString(aString);
	
	    // reduce runs of more than one space to one space only
	    newString = newString.replace(/(\s+)/g, ' '); // a space
	
	    return newString;
	},

	orderMenu: function(aMenuName, aStartIndex) {
	    if (!aStartIndex)
	        aStartIndex = 0;
	
	    var i, j;
	    var menuElement = this.getElement(aMenuName);
	    var unorderedItems = menuElement.getElementsByTagName("menuitem");

	    var orderedItems = new Array(unorderedItems.length - aStartIndex);
	    for (i = 0; i < orderedItems.length; i++) {
	        orderedItems[i] = unorderedItems[i + aStartIndex];
	    }
	
	    var tempItem;
	    
	    // Quick bubble sort - list is short so performance doesn't matter much
	    for (i = orderedItems.length - 1; i > 0; i--) {		
	        for (j = 0; j < i; j++) {
	            if (orderedItems[j].getAttribute("label") > orderedItems[j + 1].getAttribute("label")) {
	                tempItem = orderedItems[j + 1];
	                orderedItems[j + 1] = orderedItems[j];
	                orderedItems[j] = tempItem;
	            }
	        }
	    }
	    
	    for (i = 0; i < orderedItems.length; i++) {
	    	try {
		        menuElement.removeChild(orderedItems[i]);
		        menuElement.appendChild(orderedItems[i]);
		    }
		    catch (e) {
		    	this.logMessage("Problem moving '" + orderedItems[i].getAttribute("label") + "' in '" + aMenuName + "'");
		    }
	    }

		// No longer need to do this because we are actually moving the menu
		// items and not just attempting to re-order using the ordinal attribute
		//
	    // menuElement.setAttribute("hidden", "true");
	    // menuElement.setAttribute("hidden", "false");
	},

	prompt: function(aMsg, aInitial, aParent, aTitle, aDefaultValue) {
	     if (!aParent)
	         aParent = window;
	     if (!aTitle)
	         aTitle = MSG_PROMPT; // probably just in seamonkey
	     
	     var rv = { value: aInitial };
	 
	     if (!GB_PROMPT_SVC.prompt(aParent, aTitle, aMsg, rv, null, {value: null})) {
	         return aDefaultValue;
	     }
	     else {
	     	return rv.value;
	     }
	},

	setInPrefs: function(aBoolean) {
		if (aBoolean == true || aBoolean == "true") {
			this.mInPrefs = true;
		}
		else {
			this.mInPrefs = false;
		}
	},

	showWebPage: function(pageUrl) {
		var browserWindow = GB_WINDOW_SVC.getMostRecentWindow("navigator:browser");
		if (browserWindow) {
			try {
				browserWindow.delayedOpenTab(pageUrl);
			}
			catch(e) {
				browserWindow.loadURI(pageUrl);
			}
		}
		else {
			window.open(pageUrl);
		}
	},

	trimString: function(aString) {
	    if (!aString) {
	    	return "";
	    }
	    
	    return aString.replace(/(^\s+)|(\s+$)/g, '');
	},
	
	updateClass: function (aElement, aPosition, aNewClass) {
	    var oldClass = aElement.getAttribute("class");
	    var classes = oldClass.split(" ");
	    classes[aPosition] = aNewClass;
	    aElement.setAttribute("class", classes.join(" "));
	}
};

myGooglebarUtil.initialize();


var myGooglebarInstallUtil = {

	isNewInstall: function() {
		if (!this.hasBrowserWindow()) {
			return;
		}
		
		// Look for a zero byte file - googlebarIsNew
		// This file is only needed if someone changes or adds prefs
		// It puts the responsibility on the user for prefs that point to old
		// meanings. Unless someone want to make a versioned table of prefs :)
		
		var dSrvI = GB_DIR_OBJ.get("AChrom", Components.interfaces.nsIFile);
			
		dSrvI.append("googlebar");
		dSrvI.append("content");
		dSrvI.append("install");
		dSrvI.append("googlebarIsNew");

		var retv = dSrvI.exists();
		var cTyp = null;

		if (retv) {
			cTyp ="AChrom";
		} 
		else {
			dSrvI = dSrv.get("UChrm", Components.interfaces.nsIFile);

			dSrvI.append("googlebar");
			dSrvI.append("content");
			dSrvI.append("install");
			dSrvI.append("googlebarIsNew");

			retv = dSrvI.exists();

			if (retv) {
				cTyp = "UChrm";
			}
		}

		// returns our c:\\windows\stuff as file:///c:/windows
		// not used yet, but it may be later
		// var thisUri = getFileUri(dSrvI);
		
		if (!cTyp) {
			// myGooglebarUtil.logMessage("googlebarIsNew NOT Found in any chrome location\n");
			// Here we can set a pref that is really a flag
			// that then works for every tab and window.
			// that we have checked for any new installation.
			return;
		}
		else if (cTyp == "AChrom" || cTyp == "UChrm") {
			myGooglebarUtil.logMessage("Found googlebarIsNew in user profile");

			// Tell the user to check their preferences,
			// because they are probably there but have changed meaning.
			
			if (confirm(myGooglebarUtil.getProperty("gb.check4NewInstall.1"))) {
				myGooglebar.openPrefs();
				// delete the triggerfile
				dSrvI.remove(true);
			}
			else if(!confirm(myGooglebarUtil.getProperty("gb.check4NewInstall.2"))) {
				// delete the triggerfile
				dSrvI.remove(true) ;
			}
		}
		else {
			myGooglebarUtil.logMessage("Unexpected type " + cTyp + " found when checking new install");
		}
	},
	
	hasBrowserWindow: function() {
		// Quicklaunch seems to be a full browser window.
	
		if (GB_WINDOW_SVC.getMostRecentWindow("navigator:browser")) {
			if ("arguments" in window && window.arguments.length > 1 && window.arguments[1]) {
				if (window.arguments[1].indexOf("charset=") != -1) {
					var arrayArgComponents = window.arguments[1].split("=");

					if (arrayArgComponents) {
						getMarkupDocumentViewer().defaultCharacterSet = arrayArgComponents[1];
					}
				}
				else if (window.arguments[1].indexOf("turbo=yes") != -1) {
					// var gb_turboMode = true;
					return false ;
				}
			}
			
			// got a browser window besides quicklaunch
	   		return true ;
		}
		
		// no browsers
		return false;
	}
}